@@ -63,4 +63,11 @@ static inline struct nlattr *nla_find_nested(struct nlattr *nla, int attrtype)
 }
 #endif
 
+#ifndef HAVE_NLA_IS_LAST
+static inline bool nla_is_last(const struct nlattr *nla, int rem)
+{
+	return nla->nla_len == rem;
+}
+#endif
+
 #endif /* net/netlink.h */
