@@ -1,4 +1,4 @@
-/* Copyright (c) 2012 Nicira, Inc.
+/* Copyright (c) 2012, 2013, 2014 Nicira, Inc.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -13,32 +13,30 @@
  * limitations under the License. */
 
 #include <config.h>
-#include "route-table.h"
 
 #include "compiler.h"
+#include "ovs-router.h"
+#include "route-table.h"
 
 bool
-route_table_get_name(ovs_be32 ip OVS_UNUSED, char name[IFNAMSIZ] OVS_UNUSED)
+route_table_fallback_lookup(ovs_be32 ip_dst OVS_UNUSED,
+                            char output_bridge[] OVS_UNUSED,
+                            ovs_be32 *gw)
 {
-    name[0] = '\0';
+    *gw = 0;
     return false;
 }
 
-bool
-route_table_get_ifindex(ovs_be32 ip OVS_UNUSED, int *ifindex)
+uint64_t
+route_table_get_change_seq(void)
 {
-    *ifindex = 0;
-    return false;
+    return 0;
 }
 
 void
-route_table_register(void)
-{
-}
-
-void
-route_table_unregister(void)
+route_table_init(void)
 {
+    ovs_router_init();
 }
 
 void
