@@ -21,7 +21,7 @@
 #include <string.h>
 #include "dynamic-string.h"
 #include "util.h"
-#include "vlog.h"
+#include "openvswitch/vlog.h"
 
 VLOG_DEFINE_THIS_MODULE(svec);
 
