@@ -37,7 +37,8 @@
 #include "timeval.h"
 #include "transaction.h"
 #include "trigger.h"
-#include "vlog.h"
+#include "monitor.h"
+#include "openvswitch/vlog.h"
 
 VLOG_DEFINE_THIS_MODULE(ovsdb_jsonrpc_server);
 
@@ -78,8 +79,9 @@ static void ovsdb_jsonrpc_trigger_complete_done(
     struct ovsdb_jsonrpc_session *);
 
 /* Monitors. */
-static struct json *ovsdb_jsonrpc_monitor_create(
-    struct ovsdb_jsonrpc_session *, struct ovsdb *, struct json *params);
+static struct jsonrpc_msg *ovsdb_jsonrpc_monitor_create(
+    struct ovsdb_jsonrpc_session *, struct ovsdb *, struct json *params,
+    const struct json *request_id);
 static struct jsonrpc_msg *ovsdb_jsonrpc_monitor_cancel(
     struct ovsdb_jsonrpc_session *,
     struct json_array *params,
@@ -87,6 +89,9 @@ static struct jsonrpc_msg *ovsdb_jsonrpc_monitor_cancel(
 static void ovsdb_jsonrpc_monitor_remove_all(struct ovsdb_jsonrpc_session *);
 static void ovsdb_jsonrpc_monitor_flush_all(struct ovsdb_jsonrpc_session *);
 static bool ovsdb_jsonrpc_monitor_needs_flush(struct ovsdb_jsonrpc_session *);
+static struct json *ovsdb_jsonrpc_monitor_compose_update(
+    struct ovsdb_jsonrpc_monitor *monitor, bool initial);
+
 
 /* JSON-RPC database server. */
 
@@ -102,7 +107,7 @@ struct ovsdb_jsonrpc_server {
 struct ovsdb_jsonrpc_remote {
     struct ovsdb_jsonrpc_server *server;
     struct pstream *listener;   /* Listener, if passive. */
-    struct list sessions;       /* List of "struct ovsdb_jsonrpc_session"s. */
+    struct ovs_list sessions;   /* List of "struct ovsdb_jsonrpc_session"s. */
     uint8_t dscp;
 };
 
@@ -121,7 +126,7 @@ ovsdb_jsonrpc_server_create(void)
 {
     struct ovsdb_jsonrpc_server *server = xzalloc(sizeof *server);
     ovsdb_server_init(&server->up);
-    server->max_sessions = 64;
+    server->max_sessions = 330;   /* Random limit. */
     shash_init(&server->remotes);
     return server;
 }
@@ -314,20 +319,26 @@ ovsdb_jsonrpc_server_run(struct ovsdb_jsonrpc_server *svr)
     SHASH_FOR_EACH (node, &svr->remotes) {
         struct ovsdb_jsonrpc_remote *remote = node->data;
 
-        if (remote->listener && svr->n_sessions < svr->max_sessions) {
-            struct stream *stream;
-            int error;
+        if (remote->listener) {
+            if (svr->n_sessions < svr->max_sessions) {
+                struct stream *stream;
+                int error;
 
-            error = pstream_accept(remote->listener, &stream);
-            if (!error) {
-                struct jsonrpc_session *js;
-                js = jsonrpc_session_open_unreliably(jsonrpc_open(stream),
-                                                     remote->dscp);
-                ovsdb_jsonrpc_session_create(remote, js);
-            } else if (error != EAGAIN) {
-                VLOG_WARN_RL(&rl, "%s: accept failed: %s",
+                error = pstream_accept(remote->listener, &stream);
+                if (!error) {
+                    struct jsonrpc_session *js;
+                    js = jsonrpc_session_open_unreliably(jsonrpc_open(stream),
+                                                         remote->dscp);
+                    ovsdb_jsonrpc_session_create(remote, js);
+                } else if (error != EAGAIN) {
+                    VLOG_WARN_RL(&rl, "%s: accept failed: %s",
+                                 pstream_get_name(remote->listener),
+                                 ovs_strerror(error));
+                }
+            } else {
+                VLOG_WARN_RL(&rl, "%s: connection exceeded maximum (%d)",
                              pstream_get_name(remote->listener),
-                             ovs_strerror(error));
+                             svr->max_sessions);
             }
         }
 
@@ -370,7 +381,7 @@ ovsdb_jsonrpc_server_get_memory_usage(const struct ovsdb_jsonrpc_server *svr,
 /* JSON-RPC database server session. */
 
 struct ovsdb_jsonrpc_session {
-    struct list node;           /* Element in remote's sessions list. */
+    struct ovs_list node;       /* Element in remote's sessions list. */
     struct ovsdb_session up;
     struct ovsdb_jsonrpc_remote *remote;
 
@@ -668,7 +679,7 @@ ovsdb_jsonrpc_lookup_db(const struct ovsdb_jsonrpc_session *s,
     return db;
 
 error:
-    *replyp = jsonrpc_create_reply(ovsdb_error_to_json(error), request->id);
+    *replyp = jsonrpc_create_error(ovsdb_error_to_json(error), request->id);
     ovsdb_error_destroy(error);
     return NULL;
 }
@@ -746,7 +757,7 @@ ovsdb_jsonrpc_session_lock(struct ovsdb_jsonrpc_session *s,
     return jsonrpc_create_reply(result, request->id);
 
 error:
-    reply = jsonrpc_create_reply(ovsdb_error_to_json(error), request->id);
+    reply = jsonrpc_create_error(ovsdb_error_to_json(error), request->id);
     ovsdb_error_destroy(error);
     return reply;
 }
@@ -806,7 +817,7 @@ ovsdb_jsonrpc_session_unlock(struct ovsdb_jsonrpc_session *s,
     return jsonrpc_create_reply(json_object_create(), request->id);
 
 error:
-    reply = jsonrpc_create_reply(ovsdb_error_to_json(error), request->id);
+    reply = jsonrpc_create_error(ovsdb_error_to_json(error), request->id);
     ovsdb_error_destroy(error);
     return reply;
 }
@@ -836,9 +847,8 @@ ovsdb_jsonrpc_session_got_request(struct ovsdb_jsonrpc_session *s,
     } else if (!strcmp(request->method, "monitor")) {
         struct ovsdb *db = ovsdb_jsonrpc_lookup_db(s, request, &reply);
         if (!reply) {
-            reply = jsonrpc_create_reply(
-                ovsdb_jsonrpc_monitor_create(s, db, request->params),
-                request->id);
+            reply = ovsdb_jsonrpc_monitor_create(s, db, request->params,
+                                                 request->id);
         }
     } else if (!strcmp(request->method, "monitor_cancel")) {
         reply = ovsdb_jsonrpc_monitor_cancel(s, json_array(request->params),
@@ -1020,75 +1030,18 @@ ovsdb_jsonrpc_trigger_complete_done(struct ovsdb_jsonrpc_session *s)
     }
 }
 
-/* JSON-RPC database table monitors. */
-
-enum ovsdb_jsonrpc_monitor_selection {
-    OJMS_INITIAL = 1 << 0,      /* All rows when monitor is created. */
-    OJMS_INSERT = 1 << 1,       /* New rows. */
-    OJMS_DELETE = 1 << 2,       /* Deleted rows. */
-    OJMS_MODIFY = 1 << 3        /* Modified rows. */
-};
-
-/* A particular column being monitored. */
-struct ovsdb_jsonrpc_monitor_column {
-    const struct ovsdb_column *column;
-    enum ovsdb_jsonrpc_monitor_selection select;
-};
-
-/* A row that has changed in a monitored table. */
-struct ovsdb_jsonrpc_monitor_row {
-    struct hmap_node hmap_node; /* In ovsdb_jsonrpc_monitor_table.changes. */
-    struct uuid uuid;           /* UUID of row that changed. */
-    struct ovsdb_datum *old;    /* Old data, NULL for an inserted row. */
-    struct ovsdb_datum *new;    /* New data, NULL for a deleted row. */
-};
-
-/* A particular table being monitored. */
-struct ovsdb_jsonrpc_monitor_table {
-    const struct ovsdb_table *table;
-
-    /* This is the union (bitwise-OR) of the 'select' values in all of the
-     * members of 'columns' below. */
-    enum ovsdb_jsonrpc_monitor_selection select;
-
-    /* Columns being monitored. */
-    struct ovsdb_jsonrpc_monitor_column *columns;
-    size_t n_columns;
-
-    /* Contains 'struct ovsdb_jsonrpc_monitor_row's for rows that have been
-     * updated but not yet flushed to the jsonrpc connection. */
-    struct hmap changes;
-};
-
-/* A collection of tables being monitored. */
+/* Jsonrpc front end monitor. */
 struct ovsdb_jsonrpc_monitor {
-    struct ovsdb_replica replica;
     struct ovsdb_jsonrpc_session *session;
     struct ovsdb *db;
     struct hmap_node node;      /* In ovsdb_jsonrpc_session's "monitors". */
-
     struct json *monitor_id;
-    struct shash tables;     /* Holds "struct ovsdb_jsonrpc_monitor_table"s. */
+    struct ovsdb_monitor *dbmon;
+    uint64_t unflushed;         /* The first transaction that has not been
+                                       flushed to the jsonrpc remote client. */
 };
 
-static const struct ovsdb_replica_class ovsdb_jsonrpc_replica_class;
-
-struct ovsdb_jsonrpc_monitor *ovsdb_jsonrpc_monitor_find(
-    struct ovsdb_jsonrpc_session *, const struct json *monitor_id);
-static void ovsdb_jsonrpc_monitor_destroy(struct ovsdb_replica *);
-static struct json *ovsdb_jsonrpc_monitor_get_initial(
-    const struct ovsdb_jsonrpc_monitor *);
-
-static bool
-parse_bool(struct ovsdb_parser *parser, const char *name, bool default_value)
-{
-    const struct json *json;
-
-    json = ovsdb_parser_member(parser, name, OP_BOOLEAN | OP_OPTIONAL);
-    return json ? json_boolean(json) : default_value;
-}
-
-struct ovsdb_jsonrpc_monitor *
+static struct ovsdb_jsonrpc_monitor *
 ovsdb_jsonrpc_monitor_find(struct ovsdb_jsonrpc_session *s,
                            const struct json *monitor_id)
 {
@@ -1103,40 +1056,23 @@ ovsdb_jsonrpc_monitor_find(struct ovsdb_jsonrpc_session *s,
     return NULL;
 }
 
-static void
-ovsdb_jsonrpc_add_monitor_column(struct ovsdb_jsonrpc_monitor_table *mt,
-                                 const struct ovsdb_column *column,
-                                 enum ovsdb_jsonrpc_monitor_selection select,
-                                 size_t *allocated_columns)
+static bool
+parse_bool(struct ovsdb_parser *parser, const char *name, bool default_value)
 {
-    struct ovsdb_jsonrpc_monitor_column *c;
+    const struct json *json;
 
-    if (mt->n_columns >= *allocated_columns) {
-        mt->columns = x2nrealloc(mt->columns, allocated_columns,
-                                 sizeof *mt->columns);
-    }
-
-    c = &mt->columns[mt->n_columns++];
-    c->column = column;
-    c->select = select;
+    json = ovsdb_parser_member(parser, name, OP_BOOLEAN | OP_OPTIONAL);
+    return json ? json_boolean(json) : default_value;
 }
 
-static int
-compare_ovsdb_jsonrpc_monitor_column(const void *a_, const void *b_)
-{
-    const struct ovsdb_jsonrpc_monitor_column *a = a_;
-    const struct ovsdb_jsonrpc_monitor_column *b = b_;
-
-    return a->column < b->column ? -1 : a->column > b->column;
-}
-
-static struct ovsdb_error * WARN_UNUSED_RESULT
-ovsdb_jsonrpc_parse_monitor_request(struct ovsdb_jsonrpc_monitor_table *mt,
+static struct ovsdb_error * OVS_WARN_UNUSED_RESULT
+ovsdb_jsonrpc_parse_monitor_request(struct ovsdb_monitor *dbmon,
+                                    const struct ovsdb_table *table,
                                     const struct json *monitor_request,
                                     size_t *allocated_columns)
 {
-    const struct ovsdb_table_schema *ts = mt->table->schema;
-    enum ovsdb_jsonrpc_monitor_selection select;
+    const struct ovsdb_table_schema *ts = table->schema;
+    enum ovsdb_monitor_selection select;
     const struct json *columns, *select_json;
     struct ovsdb_parser parser;
     struct ovsdb_error *error;
@@ -1172,8 +1108,8 @@ ovsdb_jsonrpc_parse_monitor_request(struct ovsdb_jsonrpc_monitor_table *mt,
     } else {
         select = OJMS_INITIAL | OJMS_INSERT | OJMS_DELETE | OJMS_MODIFY;
     }
-    mt->select |= select;
 
+    ovsdb_monitor_table_add_select(dbmon, table, select);
     if (columns) {
         size_t i;
 
@@ -1192,13 +1128,13 @@ ovsdb_jsonrpc_parse_monitor_request(struct ovsdb_jsonrpc_monitor_table *mt,
             }
 
             s = columns->u.array.elems[i]->u.string;
-            column = shash_find_data(&mt->table->schema->columns, s);
+            column = shash_find_data(&table->schema->columns, s);
             if (!column) {
                 return ovsdb_syntax_error(columns, NULL, "%s is not a valid "
                                           "column name", s);
             }
-            ovsdb_jsonrpc_add_monitor_column(mt, column, select,
-                                             allocated_columns);
+            ovsdb_monitor_add_column(dbmon, table, column, select,
+                                     allocated_columns);
         }
     } else {
         struct shash_node *node;
@@ -1206,8 +1142,8 @@ ovsdb_jsonrpc_parse_monitor_request(struct ovsdb_jsonrpc_monitor_table *mt,
         SHASH_FOR_EACH (node, &ts->columns) {
             const struct ovsdb_column *column = node->data;
             if (column->index != OVSDB_COL_UUID) {
-                ovsdb_jsonrpc_add_monitor_column(mt, column, select,
-                                                 allocated_columns);
+                ovsdb_monitor_add_column(dbmon, table, column, select,
+                                         allocated_columns);
             }
         }
     }
@@ -1215,11 +1151,13 @@ ovsdb_jsonrpc_parse_monitor_request(struct ovsdb_jsonrpc_monitor_table *mt,
     return NULL;
 }
 
-static struct json *
+static struct jsonrpc_msg *
 ovsdb_jsonrpc_monitor_create(struct ovsdb_jsonrpc_session *s, struct ovsdb *db,
-                             struct json *params)
+                             struct json *params,
+                             const struct json *request_id)
 {
     struct ovsdb_jsonrpc_monitor *m = NULL;
+    struct ovsdb_monitor *dbmon = NULL;
     struct json *monitor_id, *monitor_requests;
     struct ovsdb_error *error = NULL;
     struct shash_node *node;
@@ -1243,17 +1181,16 @@ ovsdb_jsonrpc_monitor_create(struct ovsdb_jsonrpc_session *s, struct ovsdb *db,
     }
 
     m = xzalloc(sizeof *m);
-    ovsdb_replica_init(&m->replica, &ovsdb_jsonrpc_replica_class);
-    ovsdb_add_replica(db, &m->replica);
     m->session = s;
     m->db = db;
+    m->dbmon = ovsdb_monitor_create(db, m);
+    m->unflushed = 0;
     hmap_insert(&s->monitors, &m->node, json_hash(monitor_id, 0));
     m->monitor_id = json_clone(monitor_id);
-    shash_init(&m->tables);
 
     SHASH_FOR_EACH (node, json_object(monitor_requests)) {
         const struct ovsdb_table *table;
-        struct ovsdb_jsonrpc_monitor_table *mt;
+        const char *column_name;
         size_t allocated_columns;
         const struct json *mr_value;
         size_t i;
@@ -1265,10 +1202,7 @@ ovsdb_jsonrpc_monitor_create(struct ovsdb_jsonrpc_session *s, struct ovsdb *db,
             goto error;
         }
 
-        mt = xzalloc(sizeof *mt);
-        mt->table = table;
-        hmap_init(&mt->changes);
-        shash_add(&m->tables, table->schema->name, mt);
+        ovsdb_monitor_add_table(m->dbmon, table);
 
         /* Parse columns. */
         mr_value = node->data;
@@ -1278,42 +1212,50 @@ ovsdb_jsonrpc_monitor_create(struct ovsdb_jsonrpc_session *s, struct ovsdb *db,
 
             for (i = 0; i < array->n; i++) {
                 error = ovsdb_jsonrpc_parse_monitor_request(
-                    mt, array->elems[i], &allocated_columns);
+                    m->dbmon, table, array->elems[i], &allocated_columns);
                 if (error) {
                     goto error;
                 }
             }
         } else {
             error = ovsdb_jsonrpc_parse_monitor_request(
-                mt, mr_value, &allocated_columns);
+                m->dbmon, table, mr_value, &allocated_columns);
             if (error) {
                 goto error;
             }
         }
 
-        /* Check for duplicate columns. */
-        qsort(mt->columns, mt->n_columns, sizeof *mt->columns,
-              compare_ovsdb_jsonrpc_monitor_column);
-        for (i = 1; i < mt->n_columns; i++) {
-            if (mt->columns[i].column == mt->columns[i - 1].column) {
-                error = ovsdb_syntax_error(mr_value, NULL, "column %s "
-                                           "mentioned more than once",
-                                           mt->columns[i].column->name);
-                goto error;
-            }
+        column_name = ovsdb_monitor_table_check_duplicates(m->dbmon, table);
+
+        if (column_name) {
+            error = ovsdb_syntax_error(mr_value, NULL, "column %s "
+                                       "mentioned more than once",
+                                        column_name);
+            goto error;
         }
     }
 
-    return ovsdb_jsonrpc_monitor_get_initial(m);
+    dbmon = ovsdb_monitor_add(m->dbmon);
+    if (dbmon != m->dbmon) {
+        /* Found an exisiting dbmon, reuse the current one. */
+        ovsdb_monitor_remove_jsonrpc_monitor(m->dbmon, m, m->unflushed);
+        ovsdb_monitor_add_jsonrpc_monitor(dbmon, m);
+        m->dbmon = dbmon;
+    }
+
+    ovsdb_monitor_get_initial(m->dbmon);
+    json = ovsdb_jsonrpc_monitor_compose_update(m, true);
+    json = json ? json : json_object_create();
+    return jsonrpc_create_reply(json, request_id);
 
 error:
     if (m) {
-        ovsdb_remove_replica(m->db, &m->replica);
+        ovsdb_jsonrpc_monitor_destroy(m);
     }
 
     json = ovsdb_error_to_json(error);
     ovsdb_error_destroy(error);
-    return json;
+    return jsonrpc_create_error(json, request_id);
 }
 
 static struct jsonrpc_msg *
@@ -1332,7 +1274,7 @@ ovsdb_jsonrpc_monitor_cancel(struct ovsdb_jsonrpc_session *s,
             return jsonrpc_create_error(json_string_create("unknown monitor"),
                                         request_id);
         } else {
-            ovsdb_remove_replica(m->db, &m->replica);
+            ovsdb_jsonrpc_monitor_destroy(m);
             return jsonrpc_create_reply(json_object_create(), request_id);
         }
     }
@@ -1344,311 +1286,15 @@ ovsdb_jsonrpc_monitor_remove_all(struct ovsdb_jsonrpc_session *s)
     struct ovsdb_jsonrpc_monitor *m, *next;
 
     HMAP_FOR_EACH_SAFE (m, next, node, &s->monitors) {
-        ovsdb_remove_replica(m->db, &m->replica);
+        ovsdb_jsonrpc_monitor_destroy(m);
     }
 }
 
-static struct ovsdb_jsonrpc_monitor *
-ovsdb_jsonrpc_monitor_cast(struct ovsdb_replica *replica)
-{
-    ovs_assert(replica->class == &ovsdb_jsonrpc_replica_class);
-    return CONTAINER_OF(replica, struct ovsdb_jsonrpc_monitor, replica);
-}
-
-struct ovsdb_jsonrpc_monitor_aux {
-    const struct ovsdb_jsonrpc_monitor *monitor;
-    struct ovsdb_jsonrpc_monitor_table *mt;
-};
-
-/* Finds and returns the ovsdb_jsonrpc_monitor_row in 'mt->changes' for the
- * given 'uuid', or NULL if there is no such row. */
-static struct ovsdb_jsonrpc_monitor_row *
-ovsdb_jsonrpc_monitor_row_find(const struct ovsdb_jsonrpc_monitor_table *mt,
-                               const struct uuid *uuid)
-{
-    struct ovsdb_jsonrpc_monitor_row *row;
-
-    HMAP_FOR_EACH_WITH_HASH (row, hmap_node, uuid_hash(uuid), &mt->changes) {
-        if (uuid_equals(uuid, &row->uuid)) {
-            return row;
-        }
-    }
-    return NULL;
-}
-
-/* Allocates an array of 'mt->n_columns' ovsdb_datums and initializes them as
- * copies of the data in 'row' drawn from the columns represented by
- * mt->columns[].  Returns the array.
- *
- * If 'row' is NULL, returns NULL. */
-static struct ovsdb_datum *
-clone_monitor_row_data(const struct ovsdb_jsonrpc_monitor_table *mt,
-                       const struct ovsdb_row *row)
-{
-    struct ovsdb_datum *data;
-    size_t i;
-
-    if (!row) {
-        return NULL;
-    }
-
-    data = xmalloc(mt->n_columns * sizeof *data);
-    for (i = 0; i < mt->n_columns; i++) {
-        const struct ovsdb_column *c = mt->columns[i].column;
-        const struct ovsdb_datum *src = &row->fields[c->index];
-        struct ovsdb_datum *dst = &data[i];
-        const struct ovsdb_type *type = &c->type;
-
-        ovsdb_datum_clone(dst, src, type);
-    }
-    return data;
-}
-
-/* Replaces the mt->n_columns ovsdb_datums in row[] by copies of the data from
- * in 'row' drawn from the columns represented by mt->columns[]. */
-static void
-update_monitor_row_data(const struct ovsdb_jsonrpc_monitor_table *mt,
-                        const struct ovsdb_row *row,
-                        struct ovsdb_datum *data)
-{
-    size_t i;
-
-    for (i = 0; i < mt->n_columns; i++) {
-        const struct ovsdb_column *c = mt->columns[i].column;
-        const struct ovsdb_datum *src = &row->fields[c->index];
-        struct ovsdb_datum *dst = &data[i];
-        const struct ovsdb_type *type = &c->type;
-
-        if (!ovsdb_datum_equals(src, dst, type)) {
-            ovsdb_datum_destroy(dst, type);
-            ovsdb_datum_clone(dst, src, type);
-        }
-    }
-}
-
-/* Frees all of the mt->n_columns ovsdb_datums in data[], using the types taken
- * from mt->columns[], plus 'data' itself. */
-static void
-free_monitor_row_data(const struct ovsdb_jsonrpc_monitor_table *mt,
-                      struct ovsdb_datum *data)
-{
-    if (data) {
-        size_t i;
-
-        for (i = 0; i < mt->n_columns; i++) {
-            const struct ovsdb_column *c = mt->columns[i].column;
-
-            ovsdb_datum_destroy(&data[i], &c->type);
-        }
-        free(data);
-    }
-}
-
-/* Frees 'row', which must have been created from 'mt'. */
-static void
-ovsdb_jsonrpc_monitor_row_destroy(const struct ovsdb_jsonrpc_monitor_table *mt,
-                                  struct ovsdb_jsonrpc_monitor_row *row)
-{
-    if (row) {
-        free_monitor_row_data(mt, row->old);
-        free_monitor_row_data(mt, row->new);
-        free(row);
-    }
-}
-
-static bool
-ovsdb_jsonrpc_monitor_change_cb(const struct ovsdb_row *old,
-                                const struct ovsdb_row *new,
-                                const unsigned long int *changed OVS_UNUSED,
-                                void *aux_)
-{
-    struct ovsdb_jsonrpc_monitor_aux *aux = aux_;
-    const struct ovsdb_jsonrpc_monitor *m = aux->monitor;
-    struct ovsdb_table *table = new ? new->table : old->table;
-    const struct uuid *uuid = ovsdb_row_get_uuid(new ? new : old);
-    struct ovsdb_jsonrpc_monitor_row *change;
-    struct ovsdb_jsonrpc_monitor_table *mt;
-
-    if (!aux->mt || table != aux->mt->table) {
-        aux->mt = shash_find_data(&m->tables, table->schema->name);
-        if (!aux->mt) {
-            /* We don't care about rows in this table at all.  Tell the caller
-             * to skip it.  */
-            return false;
-        }
-    }
-    mt = aux->mt;
-
-    change = ovsdb_jsonrpc_monitor_row_find(mt, uuid);
-    if (!change) {
-        change = xmalloc(sizeof *change);
-        hmap_insert(&mt->changes, &change->hmap_node, uuid_hash(uuid));
-        change->uuid = *uuid;
-        change->old = clone_monitor_row_data(mt, old);
-        change->new = clone_monitor_row_data(mt, new);
-    } else {
-        if (new) {
-            update_monitor_row_data(mt, new, change->new);
-        } else {
-            free_monitor_row_data(mt, change->new);
-            change->new = NULL;
-
-            if (!change->old) {
-                /* This row was added then deleted.  Forget about it. */
-                hmap_remove(&mt->changes, &change->hmap_node);
-                free(change);
-            }
-        }
-    }
-    return true;
-}
-
-/* Returns JSON for a <row-update> (as described in RFC 7047) for 'row' within
- * 'mt', or NULL if no row update should be sent.
- *
- * The caller should specify 'initial' as true if the returned JSON is going to
- * be used as part of the initial reply to a "monitor" request, false if it is
- * going to be used as part of an "update" notification.
- *
- * 'changed' must be a scratch buffer for internal use that is at least
- * bitmap_n_bytes(mt->n_columns) bytes long. */
 static struct json *
-ovsdb_jsonrpc_monitor_compose_row_update(
-    const struct ovsdb_jsonrpc_monitor_table *mt,
-    const struct ovsdb_jsonrpc_monitor_row *row,
-    bool initial, unsigned long int *changed)
+ovsdb_jsonrpc_monitor_compose_update(struct ovsdb_jsonrpc_monitor *m,
+                                     bool initial)
 {
-    enum ovsdb_jsonrpc_monitor_selection type;
-    struct json *old_json, *new_json;
-    struct json *row_json;
-    size_t i;
-
-    type = (initial ? OJMS_INITIAL
-            : !row->old ? OJMS_INSERT
-            : !row->new ? OJMS_DELETE
-            : OJMS_MODIFY);
-    if (!(mt->select & type)) {
-        return NULL;
-    }
-
-    if (type == OJMS_MODIFY) {
-        size_t n_changes;
-
-        n_changes = 0;
-        memset(changed, 0, bitmap_n_bytes(mt->n_columns));
-        for (i = 0; i < mt->n_columns; i++) {
-            const struct ovsdb_column *c = mt->columns[i].column;
-            if (!ovsdb_datum_equals(&row->old[i], &row->new[i], &c->type)) {
-                bitmap_set1(changed, i);
-                n_changes++;
-            }
-        }
-        if (!n_changes) {
-            /* No actual changes: presumably a row changed and then
-             * changed back later. */
-            return NULL;
-        }
-    }
-
-    row_json = json_object_create();
-    old_json = new_json = NULL;
-    if (type & (OJMS_DELETE | OJMS_MODIFY)) {
-        old_json = json_object_create();
-        json_object_put(row_json, "old", old_json);
-    }
-    if (type & (OJMS_INITIAL | OJMS_INSERT | OJMS_MODIFY)) {
-        new_json = json_object_create();
-        json_object_put(row_json, "new", new_json);
-    }
-    for (i = 0; i < mt->n_columns; i++) {
-        const struct ovsdb_jsonrpc_monitor_column *c = &mt->columns[i];
-
-        if (!(type & c->select)) {
-            /* We don't care about this type of change for this
-             * particular column (but we will care about it for some
-             * other column). */
-            continue;
-        }
-
-        if ((type == OJMS_MODIFY && bitmap_is_set(changed, i))
-            || type == OJMS_DELETE) {
-            json_object_put(old_json, c->column->name,
-                            ovsdb_datum_to_json(&row->old[i],
-                                                &c->column->type));
-        }
-        if (type & (OJMS_INITIAL | OJMS_INSERT | OJMS_MODIFY)) {
-            json_object_put(new_json, c->column->name,
-                            ovsdb_datum_to_json(&row->new[i],
-                                                &c->column->type));
-        }
-    }
-
-    return row_json;
-}
-
-/* Constructs and returns JSON for a <table-updates> object (as described in
- * RFC 7047) for all the outstanding changes within 'monitor', and deletes all
- * the outstanding changes from 'monitor'.  Returns NULL if no update needs to
- * be sent.
- *
- * The caller should specify 'initial' as true if the returned JSON is going to
- * be used as part of the initial reply to a "monitor" request, false if it is
- * going to be used as part of an "update" notification. */
-static struct json *
-ovsdb_jsonrpc_monitor_compose_table_update(
-    const struct ovsdb_jsonrpc_monitor *monitor, bool initial)
-{
-    struct shash_node *node;
-    unsigned long int *changed;
-    struct json *json;
-    size_t max_columns;
-
-    max_columns = 0;
-    SHASH_FOR_EACH (node, &monitor->tables) {
-        struct ovsdb_jsonrpc_monitor_table *mt = node->data;
-
-        max_columns = MAX(max_columns, mt->n_columns);
-    }
-    changed = xmalloc(bitmap_n_bytes(max_columns));
-
-    json = NULL;
-    SHASH_FOR_EACH (node, &monitor->tables) {
-        struct ovsdb_jsonrpc_monitor_table *mt = node->data;
-        struct ovsdb_jsonrpc_monitor_row *row, *next;
-        struct json *table_json = NULL;
-
-        HMAP_FOR_EACH_SAFE (row, next, hmap_node, &mt->changes) {
-            struct json *row_json;
-
-            row_json = ovsdb_jsonrpc_monitor_compose_row_update(
-                mt, row, initial, changed);
-            if (row_json) {
-                char uuid[UUID_LEN + 1];
-
-                /* Create JSON object for transaction overall. */
-                if (!json) {
-                    json = json_object_create();
-                }
-
-                /* Create JSON object for transaction on this table. */
-                if (!table_json) {
-                    table_json = json_object_create();
-                    json_object_put(json, mt->table->schema->name, table_json);
-                }
-
-                /* Add JSON row to JSON table. */
-                snprintf(uuid, sizeof uuid, UUID_FMT, UUID_ARGS(&row->uuid));
-                json_object_put(table_json, uuid, row_json);
-            }
-
-            hmap_remove(&mt->changes, &row->hmap_node);
-            ovsdb_jsonrpc_monitor_row_destroy(mt, row);
-        }
-    }
-
-    free(changed);
-
-    return json;
+    return ovsdb_monitor_get_update(m->dbmon, initial, &m->unflushed);
 }
 
 static bool
@@ -1657,20 +1303,23 @@ ovsdb_jsonrpc_monitor_needs_flush(struct ovsdb_jsonrpc_session *s)
     struct ovsdb_jsonrpc_monitor *m;
 
     HMAP_FOR_EACH (m, node, &s->monitors) {
-        struct shash_node *node;
-
-        SHASH_FOR_EACH (node, &m->tables) {
-            struct ovsdb_jsonrpc_monitor_table *mt = node->data;
-
-            if (!hmap_is_empty(&mt->changes)) {
-                return true;
-            }
+        if (ovsdb_monitor_needs_flush(m->dbmon, m->unflushed)) {
+            return true;
         }
     }
 
     return false;
 }
 
+void
+ovsdb_jsonrpc_monitor_destroy(struct ovsdb_jsonrpc_monitor *m)
+{
+    json_destroy(m->monitor_id);
+    hmap_remove(&m->session->monitors, &m->node);
+    ovsdb_monitor_remove_jsonrpc_monitor(m->dbmon, m, m->unflushed);
+    free(m);
+}
+
 static void
 ovsdb_jsonrpc_monitor_flush_all(struct ovsdb_jsonrpc_session *s)
 {
@@ -1679,7 +1328,7 @@ ovsdb_jsonrpc_monitor_flush_all(struct ovsdb_jsonrpc_session *s)
     HMAP_FOR_EACH (m, node, &s->monitors) {
         struct json *json;
 
-        json = ovsdb_jsonrpc_monitor_compose_table_update(m, false);
+        json = ovsdb_jsonrpc_monitor_compose_update(m, false);
         if (json) {
             struct jsonrpc_msg *msg;
             struct json *params;
@@ -1690,78 +1339,3 @@ ovsdb_jsonrpc_monitor_flush_all(struct ovsdb_jsonrpc_session *s)
         }
     }
 }
-
-static void
-ovsdb_jsonrpc_monitor_init_aux(struct ovsdb_jsonrpc_monitor_aux *aux,
-                               const struct ovsdb_jsonrpc_monitor *m)
-{
-    aux->monitor = m;
-    aux->mt = NULL;
-}
-
-static struct ovsdb_error *
-ovsdb_jsonrpc_monitor_commit(struct ovsdb_replica *replica,
-                             const struct ovsdb_txn *txn,
-                             bool durable OVS_UNUSED)
-{
-    struct ovsdb_jsonrpc_monitor *m = ovsdb_jsonrpc_monitor_cast(replica);
-    struct ovsdb_jsonrpc_monitor_aux aux;
-
-    ovsdb_jsonrpc_monitor_init_aux(&aux, m);
-    ovsdb_txn_for_each_change(txn, ovsdb_jsonrpc_monitor_change_cb, &aux);
-
-    return NULL;
-}
-
-static struct json *
-ovsdb_jsonrpc_monitor_get_initial(const struct ovsdb_jsonrpc_monitor *m)
-{
-    struct ovsdb_jsonrpc_monitor_aux aux;
-    struct shash_node *node;
-    struct json *json;
-
-    ovsdb_jsonrpc_monitor_init_aux(&aux, m);
-    SHASH_FOR_EACH (node, &m->tables) {
-        struct ovsdb_jsonrpc_monitor_table *mt = node->data;
-
-        if (mt->select & OJMS_INITIAL) {
-            struct ovsdb_row *row;
-
-            HMAP_FOR_EACH (row, hmap_node, &mt->table->rows) {
-                ovsdb_jsonrpc_monitor_change_cb(NULL, row, NULL, &aux);
-            }
-        }
-    }
-    json = ovsdb_jsonrpc_monitor_compose_table_update(m, true);
-    return json ? json : json_object_create();
-}
-
-static void
-ovsdb_jsonrpc_monitor_destroy(struct ovsdb_replica *replica)
-{
-    struct ovsdb_jsonrpc_monitor *m = ovsdb_jsonrpc_monitor_cast(replica);
-    struct shash_node *node;
-
-    json_destroy(m->monitor_id);
-    SHASH_FOR_EACH (node, &m->tables) {
-        struct ovsdb_jsonrpc_monitor_table *mt = node->data;
-        struct ovsdb_jsonrpc_monitor_row *row, *next;
-
-        HMAP_FOR_EACH_SAFE (row, next, hmap_node, &mt->changes) {
-            hmap_remove(&mt->changes, &row->hmap_node);
-            ovsdb_jsonrpc_monitor_row_destroy(mt, row);
-        }
-        hmap_destroy(&mt->changes);
-
-        free(mt->columns);
-        free(mt);
-    }
-    shash_destroy(&m->tables);
-    hmap_remove(&m->session->monitors, &m->node);
-    free(m);
-}
-
-static const struct ovsdb_replica_class ovsdb_jsonrpc_replica_class = {
-    ovsdb_jsonrpc_monitor_commit,
-    ovsdb_jsonrpc_monitor_destroy
-};
