@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2011, 2012, 2013 Nicira, Inc.
+ * Copyright (c) 2011, 2012, 2013, 2014 Nicira, Inc.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -15,17 +15,17 @@
  */
 
 #include <config.h>
-
+#undef NDEBUG
+#include "odp-util.h"
 #include <stdio.h>
-
 #include "dynamic-string.h"
 #include "flow.h"
 #include "match.h"
-#include "odp-util.h"
 #include "ofp-parse.h"
 #include "ofpbuf.h"
+#include "ovstest.h"
 #include "util.h"
-#include "vlog.h"
+#include "openvswitch/vlog.h"
 
 static int
 parse_keys(bool wc_keys)
@@ -75,10 +75,11 @@ parse_keys(bool wc_keys)
             /* Convert cls_rule back to odp_key. */
             ofpbuf_uninit(&odp_key);
             ofpbuf_init(&odp_key, 0);
-            odp_flow_key_from_flow(&odp_key, &flow, flow.in_port.odp_port);
+            odp_flow_key_from_flow(&odp_key, &flow, NULL,
+                                   flow.in_port.odp_port, true);
 
             if (odp_key.size > ODPUTIL_FLOW_KEY_BYTES) {
-                printf ("too long: %"PRIuSIZE" > %d\n",
+                printf ("too long: %"PRIu32" > %d\n",
                         odp_key.size, ODPUTIL_FLOW_KEY_BYTES);
                 exit_code = 1;
             }
@@ -147,7 +148,7 @@ parse_filter(char *filter_parse)
 
     vlog_set_levels_from_string_assert("odp_util:console:dbg");
     if (filter_parse && !strncmp(filter_parse, "filter=", 7)) {
-        filter = strdup(filter_parse+7);
+        filter = xstrdup(filter_parse + 7);
         memset(&flow_filter, 0, sizeof(flow_filter));
         memset(&wc_filter, 0, sizeof(wc_filter));
 
@@ -184,8 +185,7 @@ parse_filter(char *filter_parse)
             struct minimatch minimatch;
 
             odp_flow_key_to_flow(odp_key.data, odp_key.size, &flow);
-            odp_flow_key_to_mask(odp_mask.data, odp_mask.size, &wc.masks,
-                                 &flow);
+            odp_flow_key_to_mask(odp_mask.data, odp_mask.size, &wc, &flow);
             match_init(&match, &flow, &wc);
 
             match_init(&match_filter, &flow_filter, &wc);
@@ -215,19 +215,25 @@ parse_filter(char *filter_parse)
     return 0;
 }
 
-int
-main(int argc, char *argv[])
+static void
+test_odp_main(int argc, char *argv[])
 {
+    int exit_code = 0;
+
     set_program_name(argv[0]);
     if (argc == 2 &&!strcmp(argv[1], "parse-keys")) {
-        return parse_keys(false);
+        exit_code =parse_keys(false);
     } else if (argc == 2 &&!strcmp(argv[1], "parse-wc-keys")) {
-        return parse_keys(true);
+        exit_code =parse_keys(true);
     } else if (argc == 2 && !strcmp(argv[1], "parse-actions")) {
-        return parse_actions();
+        exit_code = parse_actions();
     } else if (argc == 3 && !strcmp(argv[1], "parse-filter")) {
-        return parse_filter(argv[2]);
+        exit_code =parse_filter(argv[2]);
     } else {
         ovs_fatal(0, "usage: %s parse-keys | parse-wc-keys | parse-actions", argv[0]);
     }
+
+    exit(exit_code);
 }
+
+OVSTEST_REGISTER("test-odp", test_odp_main);
